CON

	_clkmode = xtal1 + pll16x
	_xinfreq = 5_000_000
	_stack = ($3000 + $3000 + 100) >> 2   'accomodate display memory and stack

	x_tiles = 16
	y_tiles = 12
	paramcount = 14       
	bitmap_base = $2000
	display_base = $5000
	lines = 5

	SIN_90		= $0800
	SIN_180		= $1000
	SIN_TABLE	= $E000 >> 1

VAR
	long  tv_status     '0/1/2 = off/visible/invisible           read-only
	long  tv_enable     '0/? = off/on                            write-only
	long  tv_pins       '%ppmmm = pins                           write-only
	long  tv_mode       '%ccinp = chroma,interlace,ntsc/pal,swap write-only
	long  tv_screen     'pointer to screen (words)               write-only
	long  tv_colors     'pointer to colors (longs)               write-only               
	long  tv_hc         'horizontal cells                        write-only
	long  tv_vc         'vertical cells                          write-only
	long  tv_hx         'horizontal cell expansion               write-only
	long  tv_vx         'vertical cell expansion                 write-only
	long  tv_ho         'horizontal offset                       write-only
	long  tv_vo         'vertical offset                         write-only
	long  tv_broadcast  'broadcast frequency (Hz)                write-only
	long  tv_auralcog   'aural fm cog                            write-only
	word  screen[x_tiles * y_tiles]
	long  colors
	long  gPAngles[5]
	long  gFAngles[5]
	long  gDelay[5]
	
	long  modeText

OBJ
	tv    : "tv"
	gr    : "graphics"
	philo[5] : "philosopher"

PUB main : i | dx, dy, a, mask, x, y
	'start tv
	longmove(@tv_status, @tvparams, paramcount)
	tv_screen := @screen
	tv_colors := @colors
	tv.start(@tv_status)

	' setup colors
	colors.byte[0] := $02
	colors.byte[1] := $07
	colors.byte[2] := $CA
	colors.byte[3] := $5B


	'init tile screen
	repeat dx from 0 to tv_hc - 1
		repeat dy from 0 to tv_vc - 1
			screen[dy * tv_hc + dx] := display_base >> 6 + dy + dx * tv_vc + 0 '((dy & $3F) << 10)

	'start and setup graphics
	gr.start
	gr.setup(16, 12, 128, 96, bitmap_base)


	' give some time to TV to sync
	waitcnt(cnt + CLKFREQ*2)
	' show  scrolling text
	scrollText

	' initialize random seed
	dx := cnt
	
	' start Philosopher cogs, precalculate some values
	repeat i from 0 to 4
		philo[i].start(i)
		gPAngles[i] := $1FFF * i / 5
		gFAngles[i] := $1FFF * (i * 2 + 1) / 10 - $1FFF / 5
		gDelay[i] := ?dx

	modeText := string("Random")
	i := cnt
	repeat 1800
		oneStep
		dx := cnt - i
		if dx => CLKFREQ * 15
			quit
	philo.setMode(philo#LAZY)
	modeText := string("Lazy")
	i := cnt
	repeat 1800
		oneStep
		dx := cnt - i
		if dx => CLKFREQ * 15
			quit
	philo.setMode(philo#SYNC)
	modeText := string("Sync")
	repeat
		oneStep


PRI oneStep : i | a, x, y, dx, dy, mask, count
	' main loop
	gr.clear
	
	gr.colorwidth(1, 0)
	gr.textmode(1, 1, 6, %00_00)
	gr.text(-128, 80, modeText)
	
	' paint table and bowl
	drawTable
	
	count := "0" | ("/" << 8) | ("0" << 16)
	' paint each philosopher
	repeat i from 0 to 4
		ifnot philo[i].getState
			' skip thinking philosopher - he is not in the room
			next
		if philo[i].getState == philo#HUNGRY
			count ++
		count += (1 << 16)
		'' choose color depending on philosopher's state
		gr.colorwidth(philo[i].getState+1, 2)
		'' draw philosopher on his place
		a := gPAngles[i]
		rotate(80, 0, a, @x, @y)
		gr.arc(x, y, 10, 10, 0, $1FFF >> 4, 17, 2)

	gr.colorwidth(1, 0)
	gr.textmode(2, 2, 6, %01_01)
	gr.text(0, 0, @count)
	mask := 0
	' now paint the forks
	repeat i from 0 to 4
		if philo[i].hasLeft
			' note that the fork on philosopher's left is not on the table
			mask |= |< philo[i].getLeft
			' initial fork position
			a := gPAngles[i]
			dx := 80 -8
			dy := -13
			' if philosopher has both forks, move them few times per second
			if tick(i) AND  philo[i].hasRight
				dx -= 15
				dy += 5
			' calculate position
			rotate(dx, dy, a, @x, @y)
			' paint
			drawFork(x, y, a)
		if philo[i].hasRight
			' note that the fork on philosopher's right is not on the table
			mask |= |< philo[i].getRight
			a := gPAngles[i]
			dx := 80 -8
			dy := 13
			if not tick(i) AND philo[i].hasLeft
				dx -= 15
				dy -= 5
			rotate(dx, dy, a, @x, @y)
			drawFork(x, y, a)
	' now paint the forks left on the table
	repeat i from 0 to 4
		ifnot mask & |< i
			a := gFAngles[i]
			dx := (51 * cos(a)) ~> 16
			dy := (51 * sin(a)) ~> 16
			drawFork(dx, dy, a)
	' transfer bitmap to display
	gr.copy(display_base)


PRI cos(angle)
'' Cosine
	return sin(angle + SIN_90)

PRI sin(angle) : tmp
'' Sine
	tmp := angle
	if tmp & SIN_90
		angle := -angle
	angle |= SIN_TABLE
	angle <<= 1
	angle := word[angle]
	if tmp & SIN_180
		angle := -angle
	return angle

PRI drawFork(x, y, angle) | s, c
'' Paint fork at specified position and rotate as needed
'' Uses vector drawing
	gr.colorwidth(1, 1)
	gr.vec(x, y, $100, angle, @fork_vec)
	
PRI rotate(x, y, a, aX, aY)
'' Rotate coordinate system
	a := $2000 - a
	long[aX] := ((x * cos(a)) ~> 16) + ((y * sin(a)) ~> 16)
	long[ay] := ((x * -sin(a)) ~> 16) + ((y * cos(a)) ~> 16)

PRI tick(i)
'' Tick few times per second - for movement
	return (cnt + gDelay[i]) & %1_0000_0000_0000_0000_0000_0000

PRI drawTable : angle | x, y
'' Paint table and bowl
'' Uses vector drawing
	gr.colorwidth(1, 0)
	gr.arc(0, 0, 65, 65, 0, $1FFF >> 5, 33, 2)
	angle :=  (-($2000 * 3 / 40)) & $1FFF
	gr.vec(0, 0, $100, angle, @bowl_vec)

PRI scrollText : txt | offset, y, first
'' Output scrolling text
	gr.colorwidth(1, 0)
	gr.textmode(2, 2, 6, %01_00)
	first := @text
	offset := -90
	repeat
		y := offset
		gr.clear
		txt := first
		repeat
			gr.text(0, y, txt)
			txt += strsize(txt) + 1
			ifnot byte[txt]
				quit
			y -= 32
			if y < -128
				quit
		gr.copy(display_base)
		waitcnt(cnt + CLKFREQ/40)
		offset += 2
		if (offset == 96)
			offset := 64
			first += strsize(first) + 1
			ifnot byte[first]
				quit


DAT

tvparams                
	long    0               'status
	long    1               'enable
	long    %001_0101       'pins
	long    %0000           'mode
	long    0               'screen
	long    0               'colors
	long    x_tiles         'hc
	long    y_tiles         'vc
	long    10              'hx
	long    1               'vx
	long    0               'ho
	long    0               'vo
	long    0               'broadcast
	long    0               'auralcog

fork_vec
	word    $4000 + 0
	word    0
	word    $8000 + 0
	word    10
	word    $4000 + SIN_90
	word    3
	word    $8000 + 0
	word    3
	word    $8000 + SIN_90 + SIN_180
	word    3
	word    0

bowl_vec
	word    $4000 + $2000 / 20 * 0
	word	45
	word    $8000 + $2000 / 20 * 0
	word	60
	word    $8000 + $2000 / 20 * 1
	word	60
	word    $8000 + $2000 / 20 * 2
	word	60
	word    $8000 + $2000 / 20 * 3
	word	60
	word    $8000 + $2000 / 20 * 3
	word	45
	word    $8000 + $2000 / 20 * 4
	word	45

	word    $8000 + $2000 / 20 * 4
	word	60
	word    $8000 + $2000 / 20 * 5
	word	60
	word    $8000 + $2000 / 20 * 6
	word	60
	word    $8000 + $2000 / 20 * 7
	word	60
	word    $8000 + $2000 / 20 * 7
	word	45
	word    $8000 + $2000 / 20 * 8
	word	45

	word    $8000 + $2000 / 20 * 8
	word	60
	word    $8000 + $2000 / 20 * 9
	word	60
	word    $8000 + $2000 / 20 * 10
	word	60
	word    $8000 + $2000 / 20 * 11
	word	60
	word    $8000 + $2000 / 20 * 11
	word	45
	word    $8000 + $2000 / 20 * 12
	word	45

	word    $8000 + $2000 / 20 * 12
	word	60
	word    $8000 + $2000 / 20 * 13
	word	60
	word    $8000 + $2000 / 20 * 14
	word	60
	word    $8000 + $2000 / 20 * 15
	word	60
	word    $8000 + $2000 / 20 * 15
	word	45
	word    $8000 + $2000 / 20 * 16
	word	45
	
	word    $8000 + $2000 / 20 * 16
	word	60
	word    $8000 + $2000 / 20 * 17
	word	60
	word    $8000 + $2000 / 20 * 18
	word	60
	word    $8000 + $2000 / 20 * 19
	word	60
	word    $8000 + $2000 / 20 * 19
	word	45
	word    $8000 + $2000 
	word	45
	
	word	0


text
byte 	"Five philosophers", 0
byte 	"spend their lives", 0
byte 	"thinking and eating.", 0
byte 	" ", 0
byte 	"The philosophers ", 0
byte 	"share a common dining", 0
byte 	"room where there is", 0
byte 	"a circular table ", 0
byte 	"surrounded by five", 0
byte 	"chairs, each", 0
byte 	"belonging to one", 0
byte 	"philosopher.", 0
byte 	"In the center is", 0
byte 	"a large bowl of", 0
byte 	"spaghetti, and the", 0
byte 	"table is laid with", 0
byte 	"five forks.", 0
byte 	"On feeling hungry,", 0
byte 	"a philosopher enters", 0
byte 	"the room, sits in his", 0
byte 	"own chair, and picks", 0
byte 	"up the fork on the", 0
byte 	"left of his plate.", 0
byte 	"Unfortunately, the", 0
byte 	"spaghetti is so", 0
byte 	"tangled that he needs", 0
byte 	"to pickup and use the", 0
byte 	"fork on his right as", 0
byte 	"well. When he has", 0
byte 	"finished, he puts", 0
byte 	"down both forks and", 0
byte 	"leaves the room.", 0
byte	 0
